import ast
from collections import namedtuple
from rich import print
from simpleeval import EvalWithCompoundTypes
from typing import Any, Union, List
from edsl.exceptions import (
    SurveyRuleCannotEvaluateError,
    SurveyRuleCollectionHasNoRulesAtNodeError,
    SurveyRuleRefersToFutureStateError,
    SurveyRuleReferenceInRuleToUnknownQuestionError,
    SurveyRuleSendsYouBackwardsError,
    SurveyRuleSkipLogicSyntaxError,
)
from edsl.surveys.base import EndOfSurvey
from edsl.utilities.ast_utilities import extract_variable_names
from edsl.utilities.interface import print_table_with_rich


class Rule:
    """
    The class defines a a "rule" for what question an agent should be presented next.

    The key structure of a rule is:
    - current_q: The question at which the rule is potentially applied (an index)
    - expression: A string that if evaluates to true, then next_q (an index) is next
    - next_q: The question is true
    - priority: an integer that determines which rule is applied if multiple rules apply

    The key part of this is expression specifiying the skip logic.
    An expression can be any combination of and/or/not/parenthesis etc.

            "q1 == 'yes' or q2 == 'no'"

    The skip-logic expression has to be about questions "before" the current node.
    To determine the value of the expression, we currently, we use eval.
    Eventually, we'll use the AST of the expression to make it safer.

    If multiple rules apply, the one with the highest priority is used.
    This is to deal with the fact that when we create a survey, we give the
    next question as the default.
    So when a question is added with index, it is always given a rule the next question is index + 1, but
    given a low (-1) priority.

    Only one rule should apply at each priority level.
    If there are conflicting rules, an exception is raised.
    Ideally, we'd have a way to resolve this ex ante, perhaps to traversing the implied tree
    each time a rule is added, but for now, we'll let the error emerge at run-time.

    ## Not implemented but nice to have
    We could potentially use the question pydantic models to check for rule conflicts, as
    they define the potential trees through a survey.

    We could also use the AST to check for conflicts by inspecting the types of a rule.
    For example, if we know the answer to a question is a string, we could check that
    the expression only contains string comparisons.
    This would be a lot of work.
    """

    def __init__(
        self,
        current_q: int,
        expression: str,
        next_q: Union[int, EndOfSurvey.__class__],
        question_name_to_index: dict[str, int],
        priority: int,
    ):
        """Questions are represented by int indices."""

        self.current_q = current_q
        self.expression = expression
        self.next_q = next_q
        self.priority = priority
        self.question_name_to_index = question_name_to_index

        if not next_q == EndOfSurvey and current_q > next_q:
            raise SurveyRuleSendsYouBackwardsError

        # get the AST for the expression - used to extract
        # the variables referenced in the expression
        try:
            self.ast_tree = ast.parse(self.expression)
        except SyntaxError:
            raise SurveyRuleSkipLogicSyntaxError

        # get the names of the variables in the expression
        # e.g., q1 == 'yes' -> ['q1']
        extracted_question_names = extract_variable_names(self.ast_tree)

        # make sure all the variables in the expression are known questions
        try:
            assert all([q in question_name_to_index for q in extracted_question_names])
        except AssertionError:
            raise SurveyRuleReferenceInRuleToUnknownQuestionError

        # get the indices of the questions mentioned in the expression
        self.named_questions_by_index = [
            question_name_to_index[q] for q in extracted_question_names
        ]

        # A rule should only refer to questions that have already been asked.
        # so the named questions in the expression should not be higher than the current question
        if self.named_questions_by_index:
            if max(self.named_questions_by_index) > self.current_q:
                raise SurveyRuleRefersToFutureStateError

    def to_dict(self):
        return {
            "current_q": self.current_q,
            "expression": self.expression,
            "next_q": "EndOfSurvey" if self.next_q == EndOfSurvey else self.next_q,
            "priority": self.priority,
            "question_name_to_index": self.question_name_to_index,
        }

    @classmethod
    def from_dict(self, rule_dict):
        if rule_dict["next_q"] == "EndOfSurvey":
            rule_dict["next_q"] = EndOfSurvey

        return Rule(
            current_q=rule_dict["current_q"],
            expression=rule_dict["expression"],
            next_q=rule_dict["next_q"],
            priority=rule_dict["priority"],
            question_name_to_index=rule_dict["question_name_to_index"],
        )

    def __repr__(self):
        return f'Rule(current_q={self.current_q}, expression="{self.expression}", next_q={self.next_q}, priority={self.priority}, question_name_to_index={self.question_name_to_index})'

    def __str__(self):
        return self.__repr__()

    @property
    def question_index_to_name(self):
        """Reverses the dictionary do we can look up questions by name"""
        return {v: k for k, v in self.question_name_to_index.items()}

    def show_ast_tree(self):
        """Pretty-prints the AST tree to the terminal"""
        print(
            ast.dump(
                self.ast_tree, annotate_fields=True, indent=4, include_attributes=True
            )
        )

    def evaluate(self, answers: dict[int, Any]):
        """
        Computes the value of the expression, given a dictionary of known questions answers.
        If the expression cannot be evaluated, it raises a CannotEvaluate exception.
        """

        def substitute_in_answers(expression, answers):
            "Take the dictionary of answers and substitute them into the expression"
            for var, value in answers.items():
                # If it's a string, add quotes; otherwise, just convert to string
                if isinstance(value, str):
                    replacement = f"'{value}'"
                else:
                    replacement = str(value)

                expression = expression.replace(var, replacement)
            return expression

        try:
            to_evaluate = substitute_in_answers(self.expression, answers)
            return EvalWithCompoundTypes().eval(to_evaluate)
        except Exception as e:
            print(f"Exception in evaluation: {e}")
            raise SurveyRuleCannotEvaluateError


if __name__ == "__main__":
    r = Rule(
        current_q=1,
        expression="q1 == 'yes'",
        next_q=2,
        question_name_to_index={"q1": 1},
        priority=0,
    )



import textwrap

from edsl.exceptions import QuestionSerializationError
from edsl.exceptions import QuestionCreationValidationError
from edsl.questions.Question import RegisterQuestionsMeta

# from edsl.questions.QuestionFreeText import QuestionFreeText
# registry = RegisterQuestionsMeta.get_registered_classes()
# q2c = RegisterQuestionsMeta.question_names_to_classes()


class Meta(type):
    def __repr__(cls):
        lines = "\n".join(cls.available())
        return textwrap.dedent(
            f"""\
        Available questions: 
        {lines}        
        """
        )


class QuestionBase(metaclass=Meta):
    def __new__(cls, question_type, *args, **kwargs):
        get_question_classes = RegisterQuestionsMeta.question_types_to_classes()

        subclass = get_question_classes.get(question_type, None)
        if subclass is None:
            raise ValueError(
                f"No question registered with question_type {question_type}"
            )

        # Create an instance of the selected subclass
        instance = object.__new__(subclass)
        instance.__init__(*args, **kwargs)
        return instance

    @classmethod
    def available(cls):
        return list(RegisterQuestionsMeta.question_types_to_classes().keys())


def get_question_class(question_type):
    q2c = RegisterQuestionsMeta.question_types_to_classes()
    if question_type not in q2c:
        raise ValueError(
            f"The question type, {question_type}, is not recognized. Recognied types are: {q2c.keys()}"
        )
    return q2c.get(question_type)


# all question types must be registered here
# the key is the question type
# the value is a tuple of the module name and the class name
# CLASS_REGISTRY = {
#     "budget": ("edsl.questions.QuestionBudget", "QuestionBudget"),
#     "checkbox": ("edsl.questions.QuestionCheckBox", "QuestionCheckBox"),
#     "extract": ("edsl.questions.QuestionExtract", "QuestionExtract"),
#     "free_text": ("edsl.questions.QuestionFreeText", "QuestionFreeText"),
#     "functional": ("edsl.questions.QuestionFunctional", "QuestionFunctional"),
#     "likert_five": ("edsl.questions.derived.QuestionLikertFive", "QuestionLikertFive"),
#     "linear_scale": (
#         "edsl.questions.derived.QuestionLinearScale",
#         "QuestionLinearScale",
#     ),
#     "list": ("edsl.questions.QuestionList", "QuestionList"),
#     "multiple_choice": (
#         "edsl.questions.QuestionMultipleChoice",
#         "QuestionMultipleChoice",
#     ),
#     "numerical": ("edsl.questions.QuestionNumerical", "QuestionNumerical"),
#     "rank": ("edsl.questions.QuestionRank", "QuestionRank"),
#     "top_k": ("edsl.questions.derived.QuestionTopK", "QuestionTopK"),
#     "yes_no": ("edsl.questions.derived.QuestionYesNo", "QuestionYesNo"),
# }

question_purpose = {
    "multiple_choice": "When options are known and limited",
    "free_text": "When options are unknown or unlimited",
    "checkbox": "When multiple options can be selected",
    "numerical": "When the answer is a single numerical value e.g., a float",
    "linear_scale": "When options are text, but can be ordered e.g., daily, weekly, monthly, etc.",
    "yes_no": "When the question can be fully answered with either a yes or a no",
}


if __name__ == "__main__":
    print(QuestionBase.available())

    q = QuestionBase(
        "free_text", question_text="How are you doing?", question_name="test"
    )
    results = q.run()



class JobsErrors(Exception):
    pass


class JobsRunError(JobsErrors):
    pass


class InterviewError(Exception):
    pass


class InterviewErrorPriorTaskCanceled(InterviewError):
    pass


class InterviewTimeoutError(InterviewError):
    pass



import subprocess
import pandas as pd
import tempfile
import matplotlib.pyplot as plt
import matplotlib.image as mpimg

class ResultsGGMixin:

    def ggplot2(self, ggplot_code: str, filename=None, shape="wide"):
        # Fetching DataFrame based on shape
        if shape == "long":
            df = self.sql("select * from self", shape="long")
        elif shape == "wide":
            df = self.sql("select * from self", shape="wide", remove_prefix=True)

        # Convert DataFrame to CSV format
        csv_data = df.to_csv(index=False)

        # Embed the CSV data within the R script
        csv_data_escaped = csv_data.replace("\n", "\\n").replace("'", "\\'")
        read_csv_code = f"self <- read.csv(text = '{csv_data_escaped}', sep = ',')\n"
        
        # Load ggplot2 library
        load_ggplot2 = "library(ggplot2)\n"

        # Check if a filename is provided for the plot, if not create a temporary one
        if not filename:
            filename = tempfile.mktemp(suffix=".png")

        # Combine all R script parts
        full_r_code = load_ggplot2 + read_csv_code + ggplot_code

        # Add command to save the plot to a file
        full_r_code += f'\nggsave("{filename}", plot = last_plot(), width = 6, height = 4, device = "png")'

        result = subprocess.run(['Rscript', '-'], input=full_r_code, text=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        if result.stderr:
            print("Error in R script:", result.stderr)
        else:
            self._display_plot(filename)

    def _display_plot(self, filename):
        img = mpimg.imread(filename)
        plt.imshow(img)
        plt.axis('off')
        plt.show()




from functools import partial
from itertools import chain
from edsl.report.InputOutputDataTypes import (
    CategoricalData,
    NumericalData,
    FreeTextData,
)


class ResultsFetchMixin:
    def fetch_list(self, data_type, key) -> list:
        print("Change to using _fetch_list here")
        return self._fetch_list(data_type, key)

    def _fetch_list(self, data_type, key) -> list:
        """
        For a given data type and key, return a list of values from the data.
        It uses the filtered data, not the original data.
        >>> r = Results.create_example()
        >>> r._fetch_list('answer', 'how_feeling')
        ['Bad', 'Bad', 'Great', 'Great']
        """
        returned_list = []
        for row in self.data:
            returned_list.append(row.sub_dicts[data_type].get(key, None))

        return returned_list

    def _fetch_answer_data(self, key, element_data_class):
        """Extracts data from a results object and returns in in the corresponding element_data_class."""
        short_names_dict = {}
        if key.endswith("_comment"):
            responses = self._fetch_list("answer", key)
            question = self.survey.get_question(key[:-8])
            text = "Comment on question text: " + question.question_text
        else:
            question = self.survey.get_question(key)
            question_type = question.question_type
            # if the question has options, we use them to create the element_data_class
            options = getattr(question, "question_options", None)
            responses = self._fetch_list("answer", key)
            text = question.question_text

        if hasattr(question, "short_names_dict"):
            short_names_dict = getattr(question, "short_names_dict", {})
        else:
            short_names_dict = {}

        data = {
            "text": text,
            "short_names_dict": short_names_dict,
        }

        if element_data_class == CategoricalData:
            if question_type in ["checkbox"]:
                data["responses"] = [
                    str(r) for r in list(chain.from_iterable(responses))
                ]
            else:
                data["responses"] = [str(r) for r in responses]

            data["options"] = [str(o) for o in options]

        elif element_data_class == NumericalData:
            data["responses"] = [float(x) for x in responses]
        elif element_data_class == FreeTextData:
            data["responses"] = [str(r) for r in responses]
        else:
            raise Exception(f"Not yet implemented")

        return element_data_class(**data)

    def _fetch_other_data(self, data_type, key, element_data_class):
        if element_data_class == CategoricalData:
            responses = self._fetch_list(data_type, key)
            options = list(set(responses))
            # if data_type == "agent":
            #     text = "Agent attributes:" + self[0].agent.get_original_name(key)
            # else:
            #     text = data_type + "." + key
            text = data_type + "." + key
            return CategoricalData(
                responses=[str(r) for r in responses],
                options=[str(o) for o in options],
                text=text,
            )
        else:
            raise Exception(f"Not yet implemented")

    def _fetch_element(self, data_type, key, element_data_class):
        """Extracts data from a results object and returns in in the corresponding element_data_class."""
        data_types_to_functions = {
            "answer": self._fetch_answer_data,
            "agent": partial(self._fetch_other_data, "agent"),
            "survey": partial(self._fetch_other_data, "survey"),
            "model": partial(self._fetch_other_data, "model"),
            "scenario": partial(self._fetch_other_data, "scenario"),
        }
        return data_types_to_functions[data_type](key, element_data_class)


if __name__ == "__main__":
    from edsl.results import Results 
    r = Results.example()
    r.word_cloud_plot("how_feeling_comment")



import pydot
import tempfile
from IPython.display import Image
from edsl.utilities import is_notebook
from edsl.surveys.base import RulePriority, EndOfSurvey

class SurveyFlowVisualizationMixin:

    def show_flow(self):
        "Creates an image showing the flow of users through the survey."
        # Create a graph object
        graph = pydot.Dot(graph_type='digraph')

        # Add nodes
        for index, question in enumerate(self.questions):
            graph.add_node(pydot.Node(f'Q{index}', label=f'{question.question_name}', shape='ellipse'))
 
        # Add an "EndOfSurvey" node
        graph.add_node(pydot.Node('EndOfSurvey', label='End of Survey', shape='rectangle'))

        # Add edges for normal flow
        num_questions = len(self.questions)
        for index in range(num_questions - 1):  # From Q1 to Q3
            graph.add_edge(pydot.Edge(f'Q{index}', f'Q{index+1}'))

        graph.add_edge(pydot.Edge(f'Q{num_questions-1}', 'EndOfSurvey'))

    
        relevant_rules = [rule for rule in self.rule_collection if rule.priority > RulePriority.DEFAULT.value]


        colors = ['blue', 'red', 'purple', 'orange', 'pink', 'brown', 'grey', 'cyan', 'green']
        rule_colors = {rule: colors[i % len(colors)] for i, rule in enumerate(relevant_rules)}

        for rule in relevant_rules:
            color = rule_colors[rule]
            edge_label = f'if {rule.expression}'
            source_node = f'Q{rule.current_q}'
            target_node = f'Q{rule.next_q}' if rule.next_q != EndOfSurvey else 'EndOfSurvey'
            edge = pydot.Edge(source_node, target_node, label=edge_label, color=color, fontcolor = color)           
            graph.add_edge(edge)

        with tempfile.NamedTemporaryFile(delete=False, suffix='.png') as tmp_file:
            try:
                graph.write_png(tmp_file.name)
            except FileNotFoundError:
                print("""File not found. Most likely it's because you don't have graphviz installed. Please install it and try again.
                        It's 
                        $ sudo apt-get install graphviz 
                        on Ubuntu.
                    """)
                
            if is_notebook():
                display(Image(tmp_file.name))
            else:
                import os
                import sys
                if os.name == 'nt':  # Windows
                    os.system(f'start {tmp_file.name}')
                elif os.name == 'posix':  # macOS, Linux, Unix, etc.
                    os.system(f'open {tmp_file.name}' if sys.platform == 'darwin' else f'xdg-open {tmp_file.name}')
                



import asyncio
from typing import Coroutine, List
from edsl.results import Results, Result
from edsl.jobs.JobsRunner import JobsRunner
from edsl.utilities.decorators import jupyter_nb_handler


class JobsRunnerAsyncio(JobsRunner):
    runner_name = "asyncio"

    async def run_async(
        self, n=1, verbose=False, sleep=0, debug=False, progress_bar=False
    ) -> Results:
        """Creates the tasks, runs them asynchronously, and returns the results as a Results object."""

        tasks = self._create_all_interview_tasks(self.interviews, debug)
        data = await asyncio.gather(*tasks)
        return Results(survey=self.jobs.survey, data=data)

    def _create_all_interview_tasks(self, interviews, debug) -> List[asyncio.Task]:
        """Creates an awaitable task for each interview"""
        tasks = []
        for i, interview in enumerate(interviews):
            interviewing_task = self._interview_task(interview, i, debug)
            tasks.append(asyncio.create_task(interviewing_task))
        return tasks

    async def _interview_task(self, interview, i, debug):
        # Assuming async_conduct_interview and Result are defined and work asynchronously
        answer, valid_results = await interview.async_conduct_interview(debug=debug)

        answer_key_names = {k for k in set(answer.keys()) if not k.endswith("_comment")}

        assert len(valid_results) == len(answer_key_names)

        question_name_to_prompts = dict({})
        for result in valid_results:
            question_name = result["question_name"]
            question_name_to_prompts[question_name] = {
                "user_prompt": result["prompts"]["user_prompt"],
                "system_prompt": result["prompts"]["system_prompt"],
            }

        prompt_dictionary = {}
        for answer_key_name in answer_key_names:
            prompt_dictionary[
                answer_key_name + "_user_prompt"
            ] = question_name_to_prompts[answer_key_name]["user_prompt"]
            prompt_dictionary[
                answer_key_name + "_system_prompt"
            ] = question_name_to_prompts[answer_key_name]["system_prompt"]

        # breakpoint()
        #        user_prompt = [for key in answer_key_names]
        # user_prompts, system_prompts = self._get_prompts(answer, prompt_data)
        # breakpoint()

        result = Result(
            agent=interview.agent,
            scenario=interview.scenario,
            model=interview.model,
            iteration=i,
            answer=answer,
            prompt=prompt_dictionary,
        )
        return result

    @jupyter_nb_handler
    def run(
        self, n=1, verbose=False, sleep=0, debug=False, progress_bar=False
    ) -> Coroutine:
        """Runs a collection of interviews, handling both async and sync contexts."""
        return self.run_async(n, verbose, sleep, debug, progress_bar)



from __future__ import annotations
from typing import Optional
from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice


class QuestionLikertFive(QuestionMultipleChoice):
    """
    This question asks the user to respond to a statement on a 5-point Likert scale.

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_text` is the text of the question (string)

    Optional arguments:
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionLikertFive.default_instructions`
    - `question_options` are the options the user should select from (list of strings). If not provided, the default likert options are used. To view them, run `QuestionLikertFive.likert_options`
    - `short_names_dict` maps question_options to short names (dictionary mapping strings to strings)

    For an example, see `QuestionLikertFive.example()`
    """

    question_type = "likert_five"
    likert_options: list[str] = [
        "Strongly disagree",
        "Disagree",
        "Neutral",
        "Agree",
        "Strongly agree",
    ]
    # default_instructions = QuestionMultipleChoice.default_instructions

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: Optional[list[str]] = likert_options,
        short_names_dict: Optional[dict[str, str]] = None,
    ):
        super().__init__(
            question_name=question_name,
            question_text=question_text,
            question_options=question_options,
            short_names_dict=short_names_dict,
        )

    ################
    # Helpful
    ################
    @classmethod
    def example(cls) -> QuestionLikertFive:
        return cls(
            question_name="happy_raining",
            question_text="I'm only happy when it rains.",
        )


def main():
    from edsl.questions.derived.QuestionLikertFive import QuestionLikertFive

    q = QuestionLikertFive.example()
    q.question_text
    q.question_options
    q.question_name
    q.short_names_dict
    # validate an answer
    q.validate_answer({"answer": 0, "comment": "I like custard"})
    # translate answer code
    q.translate_answer_code_to_answer(0, {})
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



import asyncio

from edsl.jobs import Jobs
from edsl.results import Results, Result
from edsl.jobs.JobsRunner import JobsRunner


class JobsRunnerDryRun(JobsRunner):
    runner_name = "dryrun"

    def __init__(self, jobs: Jobs):
        super().__init__(jobs)

    def run(
        self, n=1, verbose=False, sleep=0, debug=False, progress_bar=False
    ) -> Results:
        """Runs a collection of interviews."""

        print(f"This will run {len(self.interviews)} interviews.")



from typing import Optional, Callable
from edsl.questions import Question
from edsl.questions.descriptors import FunctionDescriptor


class QuestionFunctional(Question):
    """
    A special type of question that's *not* answered by an LLM
    - Instead, it is "answered" by a function that is passed in, `func`.
    - Useful for questions that require some kind of computation first
      or are the result of a multi-step process.
    See `compose_questions` in `compose_functions.py` for an example of how this is used.

    Notes
    - 'func' is a function that takes in a scenario and agent traits and returns an answer.
    - QuestionFunctional is not meant to be instantiated directly by end-users, but rather
      it is meant to be subclassed by us to create new function types.
    - It is probably *not* safe to allow end-users to have the ability to pass functional-derived questions.
      They could monkey-patch the function to do something malicious, e.g., to replace our function logic
      with "os.system('rm -rf /')".
    - One possible solution is to have interfaces they can pass via the API, like so:
      QuestionDropdown(question_name = "dropdown", question_options = ["a", "b", "c"]...)
      which we then translate to the real QuestionFunctional `under the hood.`

    To see how it's used, see `tests/test_QuestionFunctional_construction_from_function`
    """

    func: Callable = FunctionDescriptor()
    question_type = "functional"
    default_instructions = ""

    def __init__(
        self,
        question_name: str,
        func: Callable,
        question_text: Optional[str] = "Functional question",
    ):
        self.question_name = question_name
        self.func = func
        self.question_text = question_text
        self.instructions = self.default_instructions

    def validate_answer(self, answer: dict[str, str]):
        """Required by Question, but not used by QuestionFunctional"""
        raise NotImplementedError

    def answer_question_directly(self, scenario, agent_traits=None):
        return {"answer": self.func(scenario, agent_traits), "comment": None}

    def translate_answer_code_to_answer(self, answer, scenario):
        """Required by Question, but not used by QuestionFunctional"""
        return None

    def simulate_answer(self, human_readable=True) -> dict[str, str]:
        """Required by Question, but not used by QuestionFunctional"""
        raise NotImplementedError

    @classmethod
    def example(cls):
        """Required by Question, but not used by QuestionFunctional"""

        silly_function = lambda scenario, agent_traits: scenario.get(
            "a", 1
        ) + scenario.get("b", 2)
        return cls(
            question_name="add_two_numbers",
            question_text="functional",
            func=silly_function,
        )



from __future__ import annotations
import re
import json
from typing import Any
from edsl.questions import Question
from edsl.questions.descriptors import AnswerTemplateDescriptor
from edsl.scenarios import Scenario
from edsl.utilities import random_string


class QuestionExtract(Question):
    """
    This question asks the user to extract values from a string, and return them in a given template.

    Arguments:
    - `question_name` is the name of the question (string)
    - `question_text` is the text of the question (string)
    - `answer_template` is the template for the answer (dictionary mapping strings to strings)

    Optional arguments:
    - `instructions` are the instructions for the question (string). If not provided, the default instructions are used. To view them, run `QuestionExtract.default_instructions`

    For an example, run `QuestionExtract.example()`
    """

    question_type = "extract"
    answer_template: dict[str, Any] = AnswerTemplateDescriptor()

    def __init__(
        self,
        question_text: str,
        answer_template: dict[str, Any],
        question_name: str,
    ):
        self.question_name = question_name
        self.question_text = question_text
        self.answer_template = answer_template

    ################
    # Answer methods
    ################
    def validate_answer(self, answer: Any) -> dict[str, Any]:
        """Validates the answer"""
        # raw_json = answer["answer"]
        # fixed_json_data = re.sub(r"\'", '"', raw_json)
        # answer["answer"] = json.loads(fixed_json_data)
        self.validate_answer_template_basic(answer)
        # self.validate_answer_key_value(answer, "answer", dict)

        self.validate_answer_extract(answer)
        return answer

    def translate_answer_code_to_answer(self, answer, scenario: Scenario = None):
        """Returns the answer in a human-readable format"""
        return answer

    def simulate_answer(self, human_readable: bool = True) -> dict[str, str]:
        """Simulates a valid answer for debugging purposes"""
        return {
            "answer": {key: random_string() for key in self.answer_template.keys()},
            "comment": random_string(),
        }

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionExtract:
        return cls(
            question_name="extract_name",
            question_text="My name is Moby Dick. I have a PhD in astrology, but I'm actually a truck driver",
            answer_template={"name": "John Doe", "profession": "Carpenter"},
        )


# main
def main():
    from edsl.questions.QuestionExtract import QuestionExtract

    q = QuestionExtract.example()
    q.question_text
    q.question_name
    q.answer_template
    q.validate_answer({"answer": {"name": "Moby", "profession": "truck driver"}})
    q.translate_answer_code_to_answer(
        {"answer": {"name": "Moby", "profession": "truck driver"}}
    )
    # simulate answer
    q.simulate_answer()
    q.simulate_answer(human_readable=False)
    q.validate_answer(q.simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q



import textwrap

from edsl.prompts.QuestionInstructionsBase import QuestionInstuctionsBase


class Extract(QuestionInstuctionsBase):
    question_type = "extract"
    model = "gpt-4-1106-preview"
    default_instructions = textwrap.dedent(
        """\
        You are given the following input: "{{question_text}}".
        Create an ANSWER should be formatted like this: "{{ answer_template }}",
        and it should have the same keys but values extracted from the input.
        If the value of a key is not present in the input, fill with "null".
        Return a valid JSON formatted like this:
        {"answer": <put your ANSWER here>}
        ONLY RETURN THE JSON, AND NOTHING ELSE.
        """
    )



from abc import ABC, abstractmethod, ABCMeta
import io
import gzip
import json

from rich.console import Console
from rich.table import Table
from IPython.display import display

from edsl.utilities import is_notebook


class RichPrintingMixin:
    def for_console(self):
        with io.StringIO() as buf:
            console = Console(file=buf, record=True)
            table = self.rich_print()
            console.print(table)
            return console.export_text()

    def __str__(self):
        return self.for_console()

    def print(self):
        if is_notebook():
            display(self.rich_print())
        else:
            from rich.console import Console

            console = Console()
            console.print(self.rich_print())


class PersistenceMixin:
    def save(self, filename):
        with gzip.open(filename, "wb") as f:
            f.write(json.dumps(self.to_dict()).encode("utf-8"))

    @classmethod
    def load(cls, filename):
        with gzip.open(filename, "rb") as f:
            file_contents = f.read()
            file_contents_decoded = file_contents.decode("utf-8")
            d = json.loads(file_contents_decoded)
            # d = json.loads(f.read().decode("utf-8"))
        return cls.from_dict(d)

    def post(self):
        from edsl.utilities.pastebin import post

        post(self)


class RegisterSubclassesMeta(ABCMeta):
    _registry = {}

    def __init__(cls, name, bases, nmspc):
        super(RegisterSubclassesMeta, cls).__init__(name, bases, nmspc)
        if cls.__name__ != "Base":
            RegisterSubclassesMeta._registry[cls.__name__] = cls

    @staticmethod
    def get_registry():
        return dict(RegisterSubclassesMeta._registry)


class Base(RichPrintingMixin, PersistenceMixin, ABC, metaclass=RegisterSubclassesMeta):
    @abstractmethod
    def example():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    @abstractmethod
    def rich_print():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    @abstractmethod
    def to_dict():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    @abstractmethod
    def from_dict():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    @abstractmethod
    def code():
        """This method should be implemented by subclasses."""
        raise NotImplementedError("This method is not implemented yet.")

    def show_methods(self, show_docstrings=True):
        public_methods_with_docstrings = [
            (method, getattr(self, method).__doc__)
            for method in dir(self)
            if callable(getattr(self, method)) and not method.startswith("_")
        ]
        if show_docstrings:
            for method, documentation in public_methods_with_docstrings:
                print(f"{method}: {documentation}")
        else:
            return [x[0] for x in public_methods_with_docstrings]



from .agents import (
    AgentAttributeLookupCallbackError,
    AgentCombinationError,
    AgentLacksLLMError,
    AgentRespondedWithBadJSONError,
)
from .configuration import (
    InvalidEnvironmentVariableError,
    MissingEnvironmentVariableError,
)
from .data import (
    DatabaseConnectionError,
    DatabaseCRUDError,
    DatabaseIntegrityError,
)

from .jobs import JobsRunError, InterviewErrorPriorTaskCanceled, InterviewTimeoutError

from .language_models import (
    LanguageModelResponseNotJSONError,
    LanguageModelMissingAttributeError,
    LanguageModelAttributeTypeError,
    LanguageModelDoNotAddError,
)
from .questions import (
    QuestionAnswerValidationError,
    QuestionAttributeMissing,
    QuestionCreationValidationError,
    QuestionResponseValidationError,
    QuestionSerializationError,
    QuestionScenarioRenderError,
)
from .results import (
    ResultsBadMutationstringError,
    ResultsColumnNotFoundError,
    ResultsInvalidNameError,
    ResultsMutateError,
)
from .surveys import (
    SurveyCreationError,
    SurveyHasNoRulesError,
    SurveyRuleCannotEvaluateError,
    SurveyRuleCollectionHasNoRulesAtNodeError,
    SurveyRuleReferenceInRuleToUnknownQuestionError,
    SurveyRuleRefersToFutureStateError,
    SurveyRuleSendsYouBackwardsError,
    SurveyRuleSkipLogicSyntaxError,
)



import pandas as pd
import sqlite3
from sqlalchemy import create_engine
from enum import Enum


class SQLDataShape(Enum):
    WIDE = "wide"
    LONG = "long"


class ResultsDBMixin:
    def rows(self):
        for index, result in enumerate(self):
            yield from result.rows(index)

    def export_sql_dump(self, shape, filename):
        shape_enum = self._get_shape_enum(shape)
        conn = self.db(shape=shape_enum)

        # Open file to write SQL dump
        with open(filename, "w") as f:
            for line in conn.iterdump():
                f.write(f"{line}\n")

        # Close the connection
        conn.close()

    def backup_db_to_file(self, shape, filename):
        shape_enum = self._get_shape_enum(shape)
        # Source database connection (in-memory)
        source_conn = self.db(shape=shape_enum)

        # Destination database connection (file)
        dest_conn = sqlite3.connect(filename)

        # Backup in-memory database to file
        with source_conn:
            source_conn.backup(dest_conn)

        # Close both connections
        source_conn.close()
        dest_conn.close()

    def db(self, shape: SQLDataShape, remove_prefix=False):
        if shape == SQLDataShape.LONG:
            # Step 2: Create a SQLite Database in Memory
            conn = sqlite3.connect(":memory:")

            create_table_query = """
            CREATE TABLE self (
                id INTEGER,
                data_type TEXT,
                key TEXT, 
                value TEXT
            )
            """
            conn.execute(create_table_query)

            # # Step 3: Insert the tuples into the table
            list_of_tuples = list(self.rows())
            insert_query = (
                "INSERT INTO self (id, data_type, key, value) VALUES (?, ?, ?, ?)"
            )
            conn.executemany(insert_query, list_of_tuples)
            conn.commit()
            return conn
        elif shape == SQLDataShape.WIDE:
            db_uri = "sqlite:///:memory:"

            # Create SQLAlchemy engine with the in-memory database connection string
            engine = create_engine(db_uri)

            # Convert DataFrame to SQLite in-memory database
            df = self.to_pandas(remove_prefix=remove_prefix)
            df.to_sql("self", engine, index=False, if_exists="replace")

            # Create a connection to the SQLite database
            conn = engine.connect()
            return conn
        else:
            raise Exception("Invalid SQLDataShape")

    def _get_shape_enum(self, shape: str):
        if shape is None:
            raise Exception("Must select either 'wide' or 'long' format")
        elif shape == "wide":
            return SQLDataShape.WIDE
        elif shape == "long":
            return SQLDataShape.LONG
        else:
            raise Exception("Invalid shape: must be either 'long' or 'wide'")

    def sql(
        self,
        query: str,
        shape: str,
        remove_prefix: bool = False,
        transpose: bool = None,
        transpose_by: str = None,
        csv: bool = False,
    ):
        """Execute a SQL query and return the results as a DataFrame.
        :param query: The SQL query to execute
        :param transpose: Transpose the DataFrame if True
        :param transpose_by: Column to use as the index when transposing, otherwise the first column
        :param csv: Return the DataFrame as a CSV string if True
        """
        shape_enum = self._get_shape_enum(shape)

        conn = self.db(shape=shape_enum, remove_prefix=remove_prefix)
        df = pd.read_sql_query(query, conn)

        # Transpose the DataFrame if transpose is True
        if transpose or transpose_by:
            df = pd.DataFrame(df)
            if transpose_by:
                df = df.set_index(transpose_by)
            else:
                df = df.set_index(df.columns[0])
            df = df.transpose()

        # Return as CSV if output is "csv"
        if csv:
            return df.to_csv(index=False)
        else:
            return df

    def show_schema(self, shape: str, remove_prefix: bool = False):
        shape_enum = self._get_shape_enum(shape)
        conn = self.db(shape=shape_enum, remove_prefix=remove_prefix)

        if shape_enum == SQLDataShape.LONG:
            # Query to get the schema of all tables
            query = "SELECT type, name, sql FROM sqlite_master WHERE type='table'"

            # Execute the query
            cursor = conn.execute(query)
            schema = cursor.fetchall()

            # Close the connection
            conn.close()

            # Format and return the schema information
            schema_info = ""
            for row in schema:
                schema_info += f"Type: {row[0]}, Name: {row[1]}, SQL: {row[2]}\n"

            print(schema_info)
        elif shape_enum == SQLDataShape.WIDE:
            query = f"PRAGMA table_info(self)"
            schema = pd.read_sql(query, conn)
            print(schema)


if __name__ == "__main__":
    from edsl.results import Results

    r = Results.example()

    df = r.sql(
        "select data_type, key, value from self where data_type = 'answer'",
        shape="long",
    )
    print(df)

    df = r.sql(
        "select * from self",
        shape="wide",
    )

    df = r.sql(
        "select * from self",
        shape="wide",
    )

    r.show_schema(shape="long")

    df = r.sql(
        "select * from self",
        shape="wide",
    )

    print(df)



from docx import Document


class AgentListExportMixin:
    def docx(self) -> "Document":
        "Generates a docx document for the survey"
        doc = Document()

        doc.add_heading("EDSL Auto-Generated Agent Description")

        doc.add_paragraph(f"\n")

        for index, agent in enumerate(self.data):
            # Add question as a paragraph
            h = doc.add_paragraph()
            h.add_run(f"Agent {index + 1}").bold = True

            p = doc.add_paragraph()
            # p.add_run(agent.persona)
            for key, value in agent.traits.items():
                p.add_run(f"{key}: ").bold = True
                p.add_run(f"{value}\n")

        return doc



import json
import requests
from typing import Any, Optional, Type, Union
from edsl import CONFIG
from edsl.questions import Question
from edsl.surveys import Survey


api_url = {
    "development": "http://127.0.0.1:8000",
    "production": "https://api.goemeritus.com",
}


class Coop:
    def __init__(self, api_key: str = None, run_mode: str = None) -> None:
        self.api_key = api_key or CONFIG.EMERITUS_API_KEY
        self.run_mode = run_mode or CONFIG.EDSL_RUN_MODE

    def __repr__(self):
        return f"Client(api_key='{self.api_key}', run_mode='{self.run_mode}')"

    @property
    def headers(self) -> dict:
        return {"Authorization": f"Bearer {self.api_key}"}

    @property
    def url(self) -> str:
        return api_url[self.run_mode]

    def _send_server_request(
        self,
        uri: str,
        method: str,
        payload: Optional[dict[str, Any]] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> requests.Response:
        """Sends a request to the server and returns the response."""
        url = f"{self.url}/{uri}"

        if method.upper() in ["GET", "DELETE"]:
            response = requests.request(
                method, url, params=params, headers=self.headers
            )
        else:
            response = requests.request(method, url, json=payload, headers=self.headers)

        return response

    def _resolve_server_response(self, response: requests.Response) -> None:
        """Checks the response from the server and raises appropriate errors."""
        if response.status_code >= 400:
            raise Exception(response.json().get("detail"))

    # QUESTIONS METHODS
    def create_question(self, question: Type[Question], public: bool = False) -> dict:
        """
        Creates a Question object.
        - `question`: the EDSL Question to be sent.
        - `public`: whether the question should be public (defaults to False)
        """
        response = self._send_server_request(
            uri="api/v0/questions",
            method="POST",
            payload={"json_string": json.dumps(question.to_dict()), "public": public},
        )
        self._resolve_server_response(response)
        return response.json()

    def get_question(self, id: int) -> Type[Question]:
        """Retrieves a Question object by id."""
        response = self._send_server_request(uri=f"api/v0/questions/{id}", method="GET")
        self._resolve_server_response(response)
        return Question.from_dict(json.loads(response.json().get("json_string")))

    @property
    def questions(self) -> list[dict[str, Union[int, Question]]]:
        """Retrieves all Questions."""
        response = self._send_server_request(uri="api/v0/questions", method="GET")
        self._resolve_server_response(response)
        questions = [
            {
                "id": q.get("id"),
                "question": Question.from_dict(json.loads(q.get("json_string"))),
            }
            for q in response.json()
        ]
        return questions

    def delete_question(self, id: int) -> dict:
        """Deletes a question from the coop."""
        response = self._send_server_request(
            uri=f"api/v0/questions/{id}", method="DELETE"
        )
        self._resolve_server_response(response)
        return response.json()

    # Surveys METHODS
    def create_survey(self, survey: Type[Survey], public: bool = False) -> dict:
        """
        Creates a Question object.
        - `survey`: the EDSL Survey to be sent.
        - `public`: whether the survey should be public (defaults to False)
        """
        response = self._send_server_request(
            uri="api/v0/surveys",
            method="POST",
            payload={"json_string": json.dumps(survey.to_dict()), "public": public},
        )
        self._resolve_server_response(response)
        return response.json()

    def get_survey(self, id: int) -> Type[Survey]:
        """Retrieves a Survey object by id."""
        response = self._send_server_request(uri=f"api/v0/surveys/{id}", method="GET")
        self._resolve_server_response(response)
        return Survey.from_dict(json.loads(response.json().get("json_string")))

    @property
    def surveys(self) -> list[dict[str, Union[int, Survey]]]:
        """Retrieves all Surveys."""
        response = self._send_server_request(uri="api/v0/surveys", method="GET")
        self._resolve_server_response(response)
        surveys = [
            {
                "id": q.get("id"),
                "survey": Survey.from_dict(json.loads(q.get("json_string"))),
            }
            for q in response.json()
        ]
        return surveys

    def delete_survey(self, id: int) -> dict:
        """Deletes a Survey from the coop."""
        response = self._send_server_request(
            uri=f"api/v0/surveys/{id}", method="DELETE"
        )
        self._resolve_server_response(response)
        return response.json()


if __name__ == "__main__":
    from edsl.coop import Coop

    API_KEY = "p-llmNVgNM8pnzCWZQ6-sDCdlMgRgithISctb_9yzqU"
    RUN_MODE = "development"
    coop = Coop(api_key=API_KEY, run_mode=RUN_MODE)

    # basics
    coop
    coop.headers
    coop.url

    ##############
    # A. QUESTIONS
    ##############
    from edsl.questions import QuestionMultipleChoice
    from edsl.questions import QuestionCheckBox
    from edsl.questions import QuestionFreeText

    # check questions on server (should be an empty list)
    coop.questions
    for question in coop.questions:
        coop.delete_question(question.get("id"))

    # get a question that does not exist (should return None)
    coop.get_question(id=1)

    # now post a Question
    coop.create_question(QuestionMultipleChoice.example())
    coop.create_question(QuestionCheckBox.example(), public=False)
    coop.create_question(QuestionFreeText.example(), public=True)

    # check all questions
    coop.questions

    # or get question by id
    coop.get_question(id=1)

    # delete the question
    coop.delete_question(id=1)

    # check all questions
    coop.questions

    ##############
    # B. Surveys
    ##############
    from edsl.surveys import Survey

    # check surveys on server (should be an empty list)
    coop.surveys
    for survey in coop.surveys:
        coop.delete_survey(survey.get("id"))

    # get a survey that does not exist (should return None)
    coop.get_survey(id=1)

    # now post a Survey
    coop.create_survey(Survey.example())
    coop.create_survey(Survey.example(), public=False)
    coop.create_survey(Survey.example(), public=True)

    # check all surveys
    coop.surveys

    # or get survey by id
    coop.get_survey(id=1)

    # delete the survey
    coop.delete_survey(id=1)

    # check all surveys
    coop.surveys



import aiohttp
import json
from typing import Any
from edsl import CONFIG
from edsl.enums import LanguageModelType, InferenceServiceType
from edsl.language_models.LanguageModel import LanguageModel


def create_deep_infra_model(model_name, url, model_class_name) -> LanguageModel:
    if not LanguageModelType.is_value_valid(model_name):
        acceptable_values = [item.value for item in LanguageModelType]
        raise Exception(
            f"""
        A Prompt's model must be one of {LanguageModelType} values, which are 
        currently {acceptable_values}. You passed {model_name}."""
        )

    class LLM(LanguageModel):
        _inference_service_ = InferenceServiceType.DEEP_INFRA.value
        _model_ = model_name
        _parameters_ = {
            "temperature": 0.5,
            "top_p": 1,
            "top_k": 1,
            "max_new_tokens": 2048,
            "stopSequences": [],
            "use_cache": True,
        }
        api_token = CONFIG.get("DEEP_INFRA_API_KEY")

        async def async_execute_model_call(
            self, user_prompt: str, system_prompt: str = ""
        ) -> dict[str, Any]:
            self.url = url
            headers = {
                "Content-Type": "application/json",
                "Authorization": f"Token {self.api_token}",
            }
            data = {
                "input": f"""
                [INST]<<SYS>> {system_prompt} 
                <<SYS>[/INST]
                {user_prompt} [/INST]
                """,
                "stream": False,
                "temperature": self.temperature,
                "top_p": self.top_p,
                "top_k": self.top_k,
                "max_new_tokens": self.max_new_tokens,
            }
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    self.url, headers=headers, data=json.dumps(data)
                ) as response:
                    raw_response_text = await response.text()
                    return json.loads(raw_response_text)

        def parse_response(self, raw_response: dict[str, Any]) -> str:
            return raw_response["results"][0]["generated_text"]

    LLM.__name__ = model_class_name

    return LLM


if __name__ == "__main__":
    pass
    # results = m.execute_model_call("Why is Cape Cod a popular vacation spot?")
    # print(m.parse_response(results))



# Schemas
from edsl.questions.settings import Settings

# Base Class
from edsl.questions.Question import Question

# Core Questions
from edsl.questions.QuestionBudget import QuestionBudget
from edsl.questions.QuestionCheckBox import QuestionCheckBox
from edsl.questions.QuestionExtract import QuestionExtract
from edsl.questions.QuestionFreeText import QuestionFreeText
from edsl.questions.QuestionFunctional import QuestionFunctional
from edsl.questions.QuestionList import QuestionList
from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
from edsl.questions.QuestionNumerical import QuestionNumerical
from edsl.questions.QuestionRank import QuestionRank

# Questions derived from core questions
from edsl.questions.derived.QuestionLikertFive import QuestionLikertFive
from edsl.questions.derived.QuestionLinearScale import QuestionLinearScale
from edsl.questions.derived.QuestionTopK import QuestionTopK
from edsl.questions.derived.QuestionYesNo import QuestionYesNo

# Compose Questions
from edsl.questions.compose_questions import compose_questions



from html import escape
from IPython.display import HTML
from IPython.display import display as ipython_diplay
from PIL import Image, ImageDraw, ImageFont
from rich.console import Console
from rich.table import Table


def heartbeat_generator():
    while True:
        for c in "|/-\\":
            yield c


def gen_html_sandwich(html_inner, interactive=False):
    if interactive:
        html_header = """
            <html>
            <head>
            <script src="https://ajax.googleapis.com/ajax/libs/jquery/3.7.1/jquery.min.js"></script>
            <link rel="stylesheet" href="https://cdn.datatables.net/1.13.6/css/jquery.dataTables.css" />
            <script src="https://cdn.datatables.net/1.13.6/js/jquery.dataTables.js"></script>
            <style>
                table {
                    font-family: Arial, sans-serif;
                    border-collapse: collapse;
                    width: 100%;
                }
                
                td, th {
                    border: 1px solid #dddddd;
                    text-align: left;
                    padding: 8px;
                }
                
                tr:nth-child(even) {
                    background-color: #dddddd;
                }
            </style>
            <script>
            $(document).ready( function () {
                $('#myTable').DataTable();
            } )
            </script>
            </head>
            <body>
        """
    else:
        html_header = """
            <html>
            <head>
            <style>
                table {
                    font-family: Arial, sans-serif;
                    border-collapse: collapse;
                    width: 100%;
                }
                
                td, th {
                    border: 1px solid #dddddd;
                    text-align: left;
                    padding: 8px;
                }
                
                tr:nth-child(even) {
                    background-color: #dddddd;
                }
            </style>
            </head>
            <body>
        """

    html_footer = """
        </body>
        </html>
    """
    return html_header + html_inner + html_footer


def view_html(html):
    import tempfile
    import webbrowser

    with tempfile.NamedTemporaryFile("w", delete=False, suffix=".html") as f:
        url = "file://" + f.name
        # Write the HTML content to the file
        f.write(html)

    # Open the URL in the web browser
    webbrowser.open(url)


def human_readable_labeler_creator():
    """Creates a function that maps thread ids to human-readable labels.
    It is structured as a closure, so that the mapping is persistent.
    I.e., when the returned function is called, it will use the same
    dictionary to map thread ids to human-readable labels if it's seen that ID
    before; otherwise, it will add a new entry to the dictionary.
    This will persist across calls to the function.
    """
    d = {}

    def func(thread_id):
        if thread_id in d:
            return d[thread_id]
        else:
            d[thread_id] = len(d)
            return d[thread_id]

    return func


def get_multiline_textsize(text, font):
    # Split the text by lines
    lines = text.split("\n")

    # Initialize width and height
    max_width = 0
    total_height = 0

    for line in lines:
        # Get the size of the text for the line
        box = font.getbbox(line)
        width, height = box[2], box[3]

        # Update max_width if width of the current line is greater than max_width
        max_width = max(max_width, width)

        # Add height to total_height
        total_height += height

    return max_width, total_height


def create_image(console, image_filename):
    font_size = 15

    text = console.export_text()  # Get the console output as text.

    # Create an image from the text
    font_size = 15
    font = ImageFont.load_default()  # Use the default font to avoid file path issues.
    # text_width, text_height = ImageDraw.Draw(
    #    Image.new("RGB", (100, 100))
    # ).multiline_textsize(text, font=font)
    text_width, text_height = get_multiline_textsize(text, font)
    image = Image.new(
        "RGB", (text_width + 20, text_height + 20), color=(255, 255, 255)
    )  # Add some padding
    d = ImageDraw.Draw(image)

    # Draw text to image
    d.multiline_text((10, 10), text, font=font, fill=(0, 0, 0))
    # Save the image
    image.save(image_filename)


def display(console, table, filename):
    if filename is not None:
        with open(filename, "w") as f:
            with console.capture() as capture:
                console.print(table)
            f.write(capture.get())
        create_image(console, filename + ".png")
    else:
        console.print(table)


def print_dict_with_rich(d, key_name="Key", value_name="Value", filename=None):
    """
    >>> print_dict_with_rich({"a": 1, "b": 2, "c": 3})
    ┏━━━━━┳━━━━━━━┓
    ┃ Key ┃ Value ┃
    ┡━━━━━╇━━━━━━━┩
    │ a   │ 1     │
    │ b   │ 2     │
    │ c   │ 3     │
    └─────┴───────┘
    """
    console = Console(record=True)
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column(key_name, style="dim")
    table.add_column(value_name, style="dim")
    for key, value in d.items():
        table.add_row(key, str(value))
    display(console, table, filename)


def print_dict_as_html_table(
    d, show=False, key_name="Key", value_name="Value", filename=None
):
    """ """
    # Start the HTML table
    html_table = f'<table border="1">\n<tr><th>{escape(key_name)}</th><th>{escape(value_name)}</th></tr>\n'

    # Add rows to the HTML table
    for key, value in d.items():
        html_table += (
            f"<tr><td>{escape(str(key))}</td><td>{escape(str(value))}</td></tr>\n"
        )

    # Close the HTML table
    html_table += "</table>"

    # Print the HTML table to console
    # print(html_table)

    # Write to file if a filename is provided
    if filename:
        with open(filename, "w") as file:
            file.write(html_table)
    else:
        if show:
            view_html(gen_html_sandwich(html_table))
        else:
            return html_table


def print_list_of_dicts_with_rich(data, filename=None, split_at_dot=True):
    # Initialize console object
    """
    >>> data = [{"a": [1, 2, 3], "b": [4, 5, 6]}]
    >>> print_list_of_dicts_with_rich(data)
    ┏━━━┳━━━┓
    ┃ a ┃ b ┃
    ┡━━━╇━━━┩
    │ 1 │ 4 │
    ├───┼───┤
    │ 2 │ 5 │
    ├───┼───┤
    │ 3 │ 6 │
    └───┴───┘
    """
    console = Console(record=True)

    # Create a table object
    table = Table(show_header=True, header_style="bold magenta", show_lines=True)

    # Adding columns to the table
    for d in data:
        for key in d.keys():
            if split_at_dot:
                value = key.replace(".", "\n.")
            else:
                value = key
            table.add_column(value, style="dim")

    # Adding rows to the table
    num_rows = len(next(iter(data[0].values())))
    for i in range(num_rows):
        row = [str(d[key][i]) for d in data for key in d.keys()]
        table.add_row(*row)

    display(console, table, filename)


def print_list_of_dicts_as_html_table(data, filename=None, interactive=False):
    # Start the HTML table

    """This prints a list of dictionaries as an HTML table.
    The columns are the keys.
    The rows are the values.
    """
    html_table = '<table id="myTable" class="display">\n'
    html_table += "  <thead>\n"
    # Add the header row
    headers = [key for d in data for key in d.keys()]
    html_table += "  <tr>\n"
    for header in headers:
        html_table += f"    <th>{header}</th>\n"
    html_table += "  </tr>\n"
    html_table += "  </thead>\n</tbody>\n"

    # Determine the number of rows
    num_rows = max(len(values) for d in data for values in d.values())

    # Add the data rows
    for i in range(num_rows):
        html_table += "  <tr>\n"
        for d in data:
            for key in d.keys():
                value = d[key][i] if i < len(d[key]) else ""
                html_table += f"    <td>{value}</td>\n"
        html_table += "  </tr>\n"

    # Close the table
    html_table += "</tbody>\n"
    html_table += "</table>"

    html = gen_html_sandwich(html_table, interactive=interactive)

    # Output or save to file
    if filename:
        with open(filename, "w") as f:
            f.write(html)
    else:
        # view_html(html)
        ipython_diplay(HTML(html))


def print_public_methods_with_doc(obj):
    console = Console()
    public_methods_with_docstrings = [
        (method, getattr(obj, method).__doc__)
        for method in dir(obj)
        if callable(getattr(obj, method))
        and not method.startswith("_")
        and method != "methods"
    ]

    for method, doc in public_methods_with_docstrings:
        if doc:
            console.print(f"[bold]{method}:[/bold]", style="green")
            console.print(f"\t{doc.strip()}", style="yellow")


def print_table_with_rich(data, filename=None):
    """
    >>> data = [{"a": 1, "b": 2, "c": 3}]
    >>> print_table_with_rich(data)
    ┏━━━┳━━━┳━━━┓
    ┃ a ┃ b ┃ c ┃
    ┡━━━╇━━━╇━━━┩
    │ 1 │ 2 │ 3 │
    └───┴───┴───┘
    >>> data = [{"a": 1, "b": 2, "c": 3},{"a": 2, "b": 9, "c": 8}]
    >>> print_table_with_rich(data)
    ┏━━━┳━━━┳━━━┓
    ┃ a ┃ b ┃ c ┃
    ┡━━━╇━━━╇━━━┩
    │ 1 │ 2 │ 3 │
    │ 2 │ 9 │ 8 │
    └───┴───┴───┘
    """
    # Initialize a console object - expects a list of dictionaries
    console = Console(record=True)

    # Create a new table
    table = Table(show_header=True, header_style="bold magenta", row_styles=["", "dim"])

    # Check if data is empty; if it is, exit
    if not data:
        console.print("No data provided!")
        return

    # Add columns based on keys in the first dictionary
    for key in data[0].keys():
        table.add_column(key, style="dim")

    # Add rows to the table
    for row in data:
        table.add_row(*[str(value) for value in row.values()])

    display(console, table, filename)


if __name__ == "__main__":
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3})
    import doctest

    doctest.testmod()
    # print_list_of_dicts_with_rich([{"a": [1, 2, 3], "b": [4, 5, 6]}])
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html")
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=True)
    # print_dict_as_html_table({"a": 1, "b": 2, "c": 3}, filename="test.html", show=False)
    # print_dict_as_html_table({"a": 1, "b": 2, "c



from __future__ import annotations
import io
from collections import UserList
from typing import Optional, Union

from rich.console import Console
from rich.table import Table

from edsl.agents import Agent
from edsl.agents.AgentListExportMixin import AgentListExportMixin
from edsl.Base import Base


class AgentList(UserList, Base, AgentListExportMixin):
    def __init__(self, data: Optional[list] = None):
        if data is not None:
            super().__init__(data)
        else:
            super().__init__()

    def to(self, question_or_survey: Union["Question", "Survey"]):
        return question_or_survey.by(*self)

    def to_dict(self):
        return {"agent_list": [agent.to_dict() for agent in self.data]}

    @classmethod
    def from_dict(cls, data: dict) -> "AgentList":
        """Deserializes the dictionary back to an Agent List object."""
        agents = [Agent.from_dict(agent_dict) for agent_dict in data["agent_list"]]
        return cls(agents)

    @classmethod
    def example(cls):
        return cls([Agent.example(), Agent.example()])

    def code(self):
        lines = [
            "from edsl.agents.Agent import Agent",
            "from edsl.agents.AgentList import AgentList",
        ]
        lines.append(f"agent_list = AgentList({self.data})")
        return lines

    def rich_print(self):
        """Displays an object as a table."""
        table = Table(title="AgentList")
        table.add_column("Agents", style="bold")
        for agent in self.data:
            table.add_row(agent.rich_print())
        return table


if __name__ == "__main__":
    pass



import os
import sys
from dotenv import load_dotenv, find_dotenv
from getpass import getpass
from typing import Any
from edsl.exceptions import (
    InvalidEnvironmentVariableError,
    MissingEnvironmentVariableError,
)

CONFIG_MAP = {
    "EDSL_RUN_MODE": {
        "default": "production",
        "allowed": ["development", "production"],
        "user_message": None,
    },
    "EDSL_PASTEBIN_URL": {
        "default": "http://127.0.0.1:5000",
        "allowed": None,
        "user_message": None,
    },
    "EDSL_DATABASE_PATH": {
        "default": f"sqlite:///{os.path.join(os.getcwd(), 'edsl_cache.db')}",
        "allowed": None,
        "user_message": None,
    },
    "EDSL_LOGGING_PATH": {
        "default": f"interview.log",
        "allowed": None,
        "user_message": None,
    },
    "API_CALL_TIMEOUT_SEC": {
        "default": "60",
        "allowed": None,
        "user_message": "What is the maximum number of seconds to wait for an API call to return?",
    },
    "EMERITUS_API_KEY": {
        "default": "local",
        "allowed": None,
        "user_message": "Please provide your Emeritus API key (https://emeritus.org/).",
    },
    "OPENAI_API_KEY": {
        "default": None,
        "allowed": None,
        "user_message": "Please provide your OpenAI API key (https://platform.openai.com/api-keys).",
    },
    "DEEP_INFRA_API_KEY": {
        "default": None,
        "allowed": None,
        "user_message": "Please provide your DeepInfra API key (https://deepinfra.com/).",
    },
    "GOOGLE_API_KEY": {
        "default": None,
        "allowed": None,
        "user_message": "Please provide your Google API key (https://console.cloud.google.com/apis/credentials).",
    },
}


class Config:
    def __init__(self):
        self._load_dotenv()
        self._set_env_vars()
        self._validate_attributes()

    def _load_dotenv(self) -> None:
        """
        Loads environment variables from the .env file.
        Overrides existing env vars, unless the env var EDSL_TESTING="True".
        """
        override = True
        if os.getenv("EDSL_TESTING") == "True":
            override = False
        _ = load_dotenv(dotenv_path=find_dotenv(usecwd=True), override=override)

    def _set_env_vars(self) -> None:
        """Sets env vars as Config class attributes. If an env var is not set and has a default value in the CONFIG_MAP, sets it to the default value."""
        for env_var, config in CONFIG_MAP.items():
            # if the env var is set, set it as a CONFIG attribute as well
            if value := os.getenv(env_var):
                setattr(self, env_var, value)
            # if the env var is not set and has a default value, set it as an CONFIG attribute and as an env var
            elif default_value := config.get("default"):
                setattr(self, env_var, default_value)
                os.environ[env_var] = default_value

    def _set_env_var(self, env_var: str, config: dict[str, Any]) -> None:
        """Attempts to set an environment variable."""
        if self.EDSL_RUN_MODE == "development":
            raise MissingEnvironmentVariableError(f"Missing env var {env_var}.")
        else:
            print("=" * 50)
            print(config.get("user_message"))
            print(f"If you would like to skip this step, press enter.")
            print(f"If you would like to provide your key, do one of the following:")
            print(f"1. Set it as a regular environment variable")
            print(f"2. Create a .env file and add `{env_var}=...` to it")
            print(f"3. Enter the value below and press enter: ")
            # if the script is running in a terminal
            if sys.stdout.isatty():
                # try to use get_pass to mask the input
                try:
                    value = getpass()
                # if you fail, use input instead
                except Exception as e:
                    value = input()
            # otherwise use input
            else:
                value = input()
            value = value.strip()
            setattr(self, env_var, value)
            os.environ[env_var] = value
            if value:
                if len(value) <= 4:
                    masked_value = value
                else:
                    masked_value = value[:2] + "*" * (len(value) - 4) + value[-2:]
                print(
                    f"Environment variable {env_var} set successfully to {masked_value}."
                )
            else:
                print(f"Skipping setting environment variable {env_var}.")
            print("=" * 50)
            print("\n")

    def _validate_attributes(self):
        """Validates that all attributes are allowed values."""
        for attr, value in self.__dict__.items():
            config = CONFIG_MAP.get(attr)
            if config.get("allowed") and value not in config.get("allowed"):
                raise InvalidEnvironmentVariableError(
                    f"Variable {attr} has value {value}, which is not allowed.\n"
                    f"Allowed values are: {config.get('allowed')}. "
                )

    def get(self, env_var: str) -> str:
        """
        Returns the value of an environment variable.
        - If the environment variable is valid but not set, attempts to set it.
        """
        if env_var not in CONFIG_MAP:
            raise InvalidEnvironmentVariableError(
                f"Variable {env_var} is not a valid environment variable. "
                f"Valid environment variables are: {set(CONFIG_MAP.keys())}."
            )
        elif env_var not in self.__dict__:
            self._set_env_var(env_var, CONFIG_MAP.get(env_var))
        return self.__dict__.get(env_var)

    def show(self) -> str:
        """Prints the currently set environment vars."""
        max_env_var_length = max(len(env_var) for env_var in self.__dict__)
        print("Here are the current configuration settings:")
        for env_var, value in self.__dict__.items():
            print(f"{env_var:<{max_env_var_length}} : {value}")


# Note: Python modules are singletons. As such, once this module is imported
# the same instance of it is reused across the application.
CONFIG = Config()



from collections import UserList


class Memory(UserList):
    """This list holds the questions (stored as names) that we want the
    the agent to have available when answering a question.
    """

    def __init__(self, prior_questions: list[str] = None):
        super().__init__(prior_questions or [])

    def add_prior_question(self, prior_question):
        if prior_question not in self:
            self.append(prior_question)
        else:
            raise ValueError(f"{prior_question} is already in the memory.")

    def __repr__(self):
        return f"Memory(prior_questions={self.data})"

    def to_dict(self):
        return {"prior_questions": self}

    @classmethod
    def from_dict(cls, data):
        return cls(**data)
